﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Constants;
using Data;
using Extensions;
using Manage.DataContracts;
using Manage.Services.Contracts;
using Orchestration;
using UnityEngine;

namespace Manage.Services.Synchronizers
{
    public class PersonSynchronizer : MonoBehaviour, IPersonSynchronizer, IFamilySynchronizer, IStartupComponent
    {
        public ManageService manageService;
        public PoolController poolController;

        private Subscription _personSubscription;
        private Subscription _familySubscription;

        public Family Family { get; set; } // TODO kjellre 2017-03-06 make family synchronizer
        public List<Person> Persons { get; private set; }

        public PersonSynchronizer()
        {
            Persons = new List<Person>();
        }

        #region IStartupComponent

        public int InitOrder { get { return 4; } }
        public bool IsInitComplete { get; private set; }

        public void Initialize()
        {
            _personSubscription = manageService.Subscribe("profiles");
            _familySubscription = manageService.Subscribe("groups");

            StartCoroutine(WaitForInit());
        }

        private IEnumerator WaitForInit()
        {
            while (!_personSubscription.isReady || Persons.Count < 1 || !_familySubscription.isReady)
            {
                yield return null;
            }
            IsInitComplete = true;
        }

        #endregion

        public void AddPerson(PersonDto personDto)
        {
            var person = BuildPerson(personDto);
            Persons.Add(person);
            person.PropertyChanged += Person_PropertyChanged;
        }

        public Person BuildPerson(PersonDto personDto)
        {
            var person = new Person();
            return BuildPerson(personDto, ref person);
        }

        public Person BuildPerson(PersonDto personDto, ref Person person)
        {
            person.IsUpdateEnabled = false;
            person.Id = personDto.id;
            person.Name = personDto.name;
            if (personDto.birthDate != null)
            {
                person.DateOfBirth = personDto.birthDate.ToDateTime();
            }
            if (personDto.avatar != null)
            {
                person.Avatar = GameLogic.Instance.avatars[personDto.avatar.iconIndex];
                person.Color = GameLogic.Instance.colors[personDto.avatar.colorIndex];
            }
            person.IsUpdateEnabled = true;
            return person;
        }

        public void PersonChanged(PersonDto personDto)
        {
            var person = Persons.FirstOrDefault(p => p.Id == personDto.id);
            if (person != null)
                BuildPerson(personDto, ref person);
        }

        private void Person_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            var person = (Person)sender;
            switch (e.PropertyName)
            {
                case PersonProperties.Color:
                case PersonProperties.Avatar:
                    manageService.CallMethod("profile.update",
                        JSONObject.StringObject(person.Id),
                        JSONObject.Create(GameLogic.Instance.colors.IndexOf(person.Color)),
                        JSONObject.Create(GameLogic.Instance.avatars.IndexOf(person.Avatar)));
                    break;
                case PersonProperties.Experience:
                    manageService.CallMethod("profile.update",
                        JSONObject.Create(person.Experience));
                    break;
                default:
                    break;
            }
        }

        public void AddFamily(FamilyDto familyDto)
        {
            if (Family == null)
                Family = BuildFamily(familyDto);
        }

        private Family BuildFamily(FamilyDto familyDto)
        {
            return new Family
            {
                Name = familyDto.name,
                Id = familyDto.id
            };
        }

        #region IPersonSynchronizer
        public void InsertPerson(Person person)
        {
            var jsonObject = new JSONObject();
            jsonObject.AddField("name", person.Name);
            jsonObject.AddField("birthDate", person.DateOfBirth.ToString("yyyy-MM-dd"));
            jsonObject.AddField("iconIndex", GameLogic.Instance.avatars.IndexOf(person.Avatar));
            jsonObject.AddField("colorIndex", GameLogic.Instance.colors.IndexOf(person.Color));
            manageService.CallMethod("profile.insert",
                JSONObject.StringObject(Family.Id),
                jsonObject);
        }

        #endregion IPersonSynchronizer
    }
}
