﻿using System;
using System.Collections;
using System.Collections.Generic;
using Data;
using UI.FloatingMenus;
using UnityEngine;
using UnityEngine.EventSystems;

public class TodoPool : MonoBehaviour, IPointerDownHandler, IPointerUpHandler, IDragHandler, IPointerClickHandler
{
    private Person person;
    private Queue<GameObject> pendingTasks = new Queue<GameObject>();

    public SpriteRenderer poolBackground;
    public SpriteRenderer avatarBackround;
    public SpriteRenderer avatarForeground;

    public BoxCollider2D avatarCollider;
    public BoxCollider2D poolCollider;

    public GameObject avatarMenuPrefab;
    public GameObject PersonStatisticsPrefab;
    private bool isSpawning;
    private TaskBallController selectedTask;

    public Person Person
    {
        get { return person; }
        set
        {
            if (value != null)
            {
                person = value;
                person.PropertyChanged += PersonPropertyChanged;
                avatarBackround.color = person.Color;
                avatarForeground.sprite = person.Avatar;
            }
            else
            {
                throw new ArgumentNullException("person", "Person cannot be null!");
            }
        }
    }

    // This function is called when the MonoBehaviour will be destroyed
    public void OnDestroy()
    {
        if (person != null)
            person.PropertyChanged -= PersonPropertyChanged;
    }

    public void OnTriggerStay2D(Collider2D other)
    {
        var taskController = other.GetComponent<TaskBallController>();
        if (taskController != null && taskController.Task != null)
        {
            var task = taskController.Task;
            // Check if task ball is completely inside the pool.
            if (poolCollider.bounds.Contains(other.bounds.max) && poolCollider.bounds.Contains(other.bounds.min))
            {
                if (task.Person != person)
                    task.Person = person;
            }
            else if (!task.IsCompleted)
                task.Person = null;
        }
    }

    private void PersonPropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
    {
        switch (e.PropertyName)
        {
            case "Color":
                avatarBackround.color = person.Color;
                break;
            case "Avatar":
                avatarForeground.sprite = person.Avatar;
                break;
            default: // Do nothing
                break;
        }
    }

    private IEnumerator SpawnTasksInPool()
    {
        isSpawning = true;
        var x = poolCollider.bounds.min.x;
        var y = poolCollider.bounds.min.y;
        float nextY = y;
        while (pendingTasks.Count != 0)
        {
            var task = pendingTasks.Dequeue().GetComponent<TaskBallController>();
            task.Task.IsUpdateEnabled = false;
            task.Task.HasSettled = true;

            float radius = task.GetComponent<CircleCollider2D>().bounds.extents.x;
            float diameter = radius * 2;
            if (y + diameter * 1.2f > poolCollider.bounds.max.y)
                break;
            if (x + diameter > poolCollider.bounds.max.x)
            {
                y = nextY;
                x = poolCollider.bounds.min.x;
            }
            task.transform.position = new Vector2(x + radius, y + radius);
            task.GetComponent<TaskBallController>().EnablePhysics();
            x += diameter;
            nextY = Mathf.Max(nextY, y + diameter);

            task.EnablePhysics();
            yield return new WaitForSeconds(0.02f);
            task.Task.IsUpdateEnabled = true;
        }
        isSpawning = false;
    }

    public void AddTask(GameObject go)
    {
        pendingTasks.Enqueue(go);
        if (!isSpawning)
        {
            StartCoroutine(SpawnTasksInPool());
        }
    }

    public void OnPointerClick(PointerEventData eventData)
    {
        var colliders = Physics2D.OverlapPointAll(Camera.main.ScreenToWorldPoint(Input.mousePosition));
        bool isOpenPersonStats = false;
        foreach (var collider in colliders)
        {
            if (collider == avatarCollider)
                isOpenPersonStats = true;
            else if (collider.GetComponent<TaskBallController>() != null)
            {
                isOpenPersonStats = false;
                break;
            }
        }
        if (isOpenPersonStats)
        {
            var prefab = Instantiate(PersonStatisticsPrefab);
            prefab.transform.SetParent(GameObject.Find("Canvas").transform);
            prefab.transform.localScale = Vector3.one;
            var personStatistics = prefab.GetComponent<PersonStatistics>();
            personStatistics.Person = person;
        }
    }

    public void OnPointerDown(PointerEventData eventData)
    {
        selectedTask = GetTask();
        if (selectedTask != null)
        {
            eventData.selectedObject = selectedTask.gameObject;
            selectedTask.OnPointerDown(eventData);
        }
    }

    public void OnPointerUp(PointerEventData eventData)
    {
        var task = (selectedTask != null) ? selectedTask : GetTask();
        if (task != null)
        {
            eventData.selectedObject = task.gameObject;
            task.OnPointerUp(eventData);
            selectedTask = null;
        }
    }

    public void OnDrag(PointerEventData eventData)
    {
        var task = (selectedTask != null) ? selectedTask : GetTask();
        if (task != null)
        {
            eventData.selectedObject = task.gameObject;
            task.OnDrag(eventData);
        }
    }

    private static TaskBallController GetTask()
    {
        var colliders = Physics2D.OverlapPointAll(Camera.main.ScreenToWorldPoint(Input.mousePosition));
        TaskBallController task = null;
        foreach (var collider in colliders)
        {
            if ((task = collider.GetComponent<TaskBallController>()) != null)
                break;
        }
        return task;
    }
}
